var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import { createElement as _createElement } from "react";
import { jsx as _jsx, jsxs as _jsxs } from "react/jsx-runtime";
import { isNull, isUndefined } from 'lodash';
import React, { useState, useCallback, useMemo, useRef, useImperativeHandle } from 'react';
import { useEventCallback, useId } from '@react-devui/hooks';
import { CloseOutlined, LoadingOutlined, PlusOutlined } from '@react-devui/icons';
import { findNested, getClassName, getVerticalSidePosition } from '@react-devui/utils';
import { useGeneralContext, useDValue } from '../../hooks';
import { cloneHTMLElement, registerComponentMate, TTANSITION_DURING_POPUP, WINDOW_SPACE } from '../../utils';
import { DComboboxKeyboard } from '../_keyboard';
import { DSelectbox } from '../_selectbox';
import { DTransition } from '../_transition';
import { DCheckbox } from '../checkbox';
import { DDropdown } from '../dropdown';
import { useFormControl } from '../form';
import { useComponentConfig, usePrefixConfig, useTranslation } from '../root';
import { DTag } from '../tag';
import { DVirtualScroll } from '../virtual-scroll';
const IS_CREATE = Symbol();
const { COMPONENT_NAME } = registerComponentMate({ COMPONENT_NAME: 'DSelect' });
function Select(props, ref) {
    var _a;
    const _b = useComponentConfig(COMPONENT_NAME, props), { dRef, dFormControl, dList, dModel, dVisible, dInitialVisible = false, dPlaceholder, dSize, dLoading = false, dSearchable = false, dSearchValue, dClearable = false, dDisabled = false, dMultiple = false, dMonospaced = true, dVirtual = false, dCustomItem, dCustomSelected, dCustomSearch, dCreateItem, dPopupClassName, dInputRender, onModelChange, onVisibleChange, onClear, onSearchValueChange, onCreateItem, onScrollBottom, afterVisibleChange } = _b, restProps = __rest(_b, ["dRef", "dFormControl", "dList", "dModel", "dVisible", "dInitialVisible", "dPlaceholder", "dSize", "dLoading", "dSearchable", "dSearchValue", "dClearable", "dDisabled", "dMultiple", "dMonospaced", "dVirtual", "dCustomItem", "dCustomSelected", "dCustomSearch", "dCreateItem", "dPopupClassName", "dInputRender", "onModelChange", "onVisibleChange", "onClear", "onSearchValueChange", "onCreateItem", "onScrollBottom", "afterVisibleChange"]);
    //#region Context
    const dPrefix = usePrefixConfig();
    const { gSize, gDisabled } = useGeneralContext();
    //#endregion
    //#region Ref
    const boxRef = useRef(null);
    const popupRef = useRef(null);
    const vsRef = useRef(null);
    //#endregion
    const [t] = useTranslation();
    const uniqueId = useId();
    const listId = `${dPrefix}select-list-${uniqueId}`;
    const getGroupId = (val) => `${dPrefix}select-group-${val}-${uniqueId}`;
    const getItemId = (val) => `${dPrefix}select-item-${val}-${uniqueId}`;
    const itemsMap = useMemo(() => {
        const items = new Map();
        const reduceArr = (arr) => {
            for (const item of arr) {
                items.set(item.value, item);
                if (item.children) {
                    reduceArr(item.children);
                }
            }
        };
        reduceArr(dList);
        return items;
    }, [dList]);
    const [searchValue, changeSearchValue] = useDValue('', dSearchValue, onSearchValueChange);
    const canSelectItem = useCallback((item) => !item.disabled && !item.children, []);
    const [focusVisible, setFocusVisible] = useState(false);
    const [visible, changeVisible] = useDValue(dInitialVisible, dVisible, onVisibleChange);
    const formControlInject = useFormControl(dFormControl);
    const [_select, changeSelect] = useDValue(dMultiple ? [] : null, dModel, (value) => {
        if (onModelChange) {
            if (dMultiple) {
                onModelChange(value, value.map((v) => itemsMap.get(v)));
            }
            else {
                onModelChange(value, isNull(value) ? null : itemsMap.get(value));
            }
        }
    }, undefined, formControlInject);
    const select = useMemo(() => (dMultiple ? new Set(_select) : _select), [_select, dMultiple]);
    const size = dSize !== null && dSize !== void 0 ? dSize : gSize;
    const disabled = (_a = (dDisabled || gDisabled || (dFormControl === null || dFormControl === void 0 ? void 0 : dFormControl.control.disabled))) !== null && _a !== void 0 ? _a : false;
    const hasSearch = searchValue.length > 0;
    const hasSelected = dMultiple ? select.size > 0 : !isNull(select);
    const _filterFn = dCustomSearch === null || dCustomSearch === void 0 ? void 0 : dCustomSearch.filter;
    const filterFn = useCallback((item) => {
        const defaultFilterFn = (item) => {
            return item.label.includes(searchValue);
        };
        return _filterFn ? _filterFn(searchValue, item) : defaultFilterFn(item);
    }, [_filterFn, searchValue]);
    const sortFn = dCustomSearch === null || dCustomSearch === void 0 ? void 0 : dCustomSearch.sort;
    const searchList = useMemo(() => {
        if (!hasSearch) {
            return [];
        }
        let createItem = dCreateItem === null || dCreateItem === void 0 ? void 0 : dCreateItem(searchValue);
        if (createItem) {
            createItem = Object.assign(Object.assign({}, createItem), { [IS_CREATE]: true });
        }
        let searchList = [];
        dList.forEach((item) => {
            if (!item.children) {
                if (createItem && item.value === createItem.value) {
                    createItem = undefined;
                }
                if (filterFn(item)) {
                    searchList.push(item);
                }
            }
            else {
                const groupList = [];
                item.children.forEach((groupItem) => {
                    if (createItem && groupItem.value === createItem.value) {
                        createItem = undefined;
                    }
                    if (filterFn(groupItem)) {
                        groupList.push(groupItem);
                    }
                });
                searchList = searchList.concat(groupList);
            }
        });
        if (sortFn) {
            searchList.sort(sortFn);
        }
        if (createItem) {
            searchList.unshift(createItem);
        }
        return searchList;
    }, [dCreateItem, dList, filterFn, hasSearch, searchValue, sortFn]);
    const list = hasSearch ? searchList : dList;
    const [_noSearchFocusItem, setNoSearchFocusItem] = useState();
    const noSearchFocusItem = (() => {
        let focusItem;
        if (_noSearchFocusItem) {
            focusItem = itemsMap.get(_noSearchFocusItem.value);
            if (focusItem && canSelectItem(focusItem)) {
                return focusItem;
            }
        }
        if (hasSelected) {
            focusItem = findNested(dList, (item) => canSelectItem(item) && dMultiple ? select.has(item.value) : select === item.value);
        }
        if (isUndefined(focusItem)) {
            focusItem = findNested(dList, (item) => canSelectItem(item));
        }
        return focusItem;
    })();
    const [_searchFocusItem, setSearchFocusItem] = useState();
    const searchFocusItem = (() => {
        if (_searchFocusItem && findNested(searchList, (item) => canSelectItem(item) && item.value === _searchFocusItem.value)) {
            return _searchFocusItem;
        }
        if (hasSearch) {
            return findNested(searchList, (item) => canSelectItem(item));
        }
    })();
    const focusItem = hasSearch ? searchFocusItem : noSearchFocusItem;
    const changeFocusItem = (item) => {
        if (!isUndefined(item)) {
            hasSearch ? setSearchFocusItem(item) : setNoSearchFocusItem(item);
        }
    };
    const createItem = (item) => {
        if (!isUndefined(item)) {
            const newItem = Object.assign({}, item);
            delete newItem[IS_CREATE];
            onCreateItem === null || onCreateItem === void 0 ? void 0 : onCreateItem(newItem);
        }
    };
    const changeSelectByClick = (val) => {
        if (dMultiple) {
            changeSelect((draft) => {
                const index = draft.findIndex((v) => v === val);
                if (index !== -1) {
                    draft.splice(index, 1);
                }
                else {
                    draft.push(val);
                }
            });
        }
        else {
            changeSelect(val);
            changeVisible(false);
        }
    };
    const [popupPositionStyle, setPopupPositionStyle] = useState({
        top: '-200vh',
        left: '-200vw',
    });
    const [transformOrigin, setTransformOrigin] = useState();
    const updatePosition = useEventCallback(() => {
        if (visible && boxRef.current && popupRef.current) {
            if (dMonospaced) {
                const width = Math.min(boxRef.current.offsetWidth, window.innerWidth - WINDOW_SPACE * 2);
                const height = popupRef.current.offsetHeight;
                const { top, left, transformOrigin } = getVerticalSidePosition(boxRef.current, { width, height }, {
                    placement: 'bottom',
                    inWindow: WINDOW_SPACE,
                });
                setPopupPositionStyle({
                    top,
                    left,
                    width,
                });
                setTransformOrigin(transformOrigin);
            }
            else {
                const boxWidth = boxRef.current.offsetWidth;
                const height = popupRef.current.offsetHeight;
                const maxWidth = window.innerWidth - WINDOW_SPACE * 2;
                const width = Math.min(Math.max(popupRef.current.scrollWidth, boxWidth), maxWidth);
                const { top, left, transformOrigin } = getVerticalSidePosition(boxRef.current, { width, height }, {
                    placement: 'bottom-left',
                    inWindow: WINDOW_SPACE,
                });
                setPopupPositionStyle({
                    top,
                    left,
                    minWidth: Math.min(boxWidth, maxWidth),
                    maxWidth,
                });
                setTransformOrigin(transformOrigin);
            }
        }
    });
    useImperativeHandle(ref, () => ({
        updatePosition,
    }), [updatePosition]);
    const [selectedNode, suffixNode, selectedLabel] = (() => {
        let selectedNode = null;
        let suffixNode = null;
        let selectedLabel;
        if (dMultiple) {
            const selectedItems = [];
            for (const v of _select) {
                const item = itemsMap.get(v);
                if (item) {
                    selectedItems.push(item);
                }
                else {
                    console.warn(`Can't find item that value field is ${v}!`);
                }
            }
            suffixNode = (_jsx(DDropdown, Object.assign({ dList: selectedItems.map((item) => {
                    const { label: itemLabel, value: itemValue, disabled: itemDisabled } = item;
                    const text = dCustomSelected ? dCustomSelected(item) : itemLabel;
                    return {
                        id: itemValue,
                        label: text,
                        type: 'item',
                        disabled: itemDisabled,
                    };
                }), onItemClick: (id) => {
                    changeSelectByClick(id);
                    return false;
                } }, { children: _jsx(DTag, Object.assign({ className: `${dPrefix}select__multiple-count`, tabIndex: -1, dSize: size }, { children: select.size })) })));
            selectedNode = selectedItems.map((item) => (_jsxs(DTag, Object.assign({ className: `${dPrefix}select__multiple-tag`, dSize: size }, { children: [dCustomSelected ? dCustomSelected(item) : item.label, !(item.disabled || disabled) && (_jsx("div", Object.assign({ className: `${dPrefix}select__close`, role: "button", "aria-label": t('Close'), onClick: (e) => {
                            e.stopPropagation();
                            changeSelectByClick(item.value);
                        } }, { children: _jsx(CloseOutlined, {}) })))] }), item.value)));
        }
        else {
            if (!isNull(select)) {
                const item = itemsMap.get(select);
                if (item) {
                    selectedLabel = item.label;
                    selectedNode = dCustomSelected ? dCustomSelected(item) : selectedLabel;
                }
                else {
                    console.warn(`Can't find item that value field is ${select}!`);
                }
            }
        }
        return [selectedNode, suffixNode, selectedLabel];
    })();
    const vsPerformance = useMemo(() => ({
        dList: list,
        dItemSize: 32,
        dItemNested: (item) => ({
            list: item.children,
            emptySize: 32,
            inAriaSetsize: false,
        }),
        dItemKey: (item) => item.value,
        dFocusable: canSelectItem,
    }), [canSelectItem, list]);
    return (_jsx(DSelectbox, Object.assign({}, restProps, { onClick: (e) => {
            var _a;
            (_a = restProps.onClick) === null || _a === void 0 ? void 0 : _a.call(restProps, e);
            changeVisible((draft) => (dSearchable ? true : !draft));
        }, dRef: {
            box: boxRef,
            input: dRef === null || dRef === void 0 ? void 0 : dRef.input,
        }, dClassNamePrefix: "select", dFormControl: dFormControl, dVisible: visible, dContent: hasSelected && selectedNode, dContentTitle: selectedLabel, dPlaceholder: dPlaceholder, dSuffix: suffixNode, dSize: size, dLoading: dLoading, dSearchable: dSearchable, dClearable: dClearable, dDisabled: disabled, dUpdatePosition: {
            fn: updatePosition,
            popupRef,
            extraScrollRefs: [],
        }, dInputRender: (el) => (_jsx(DComboboxKeyboard, Object.assign({ dVisible: visible, dEditable: dSearchable, dHasSub: false, onVisibleChange: changeVisible, onFocusChange: (key) => {
                var _a, _b, _c, _d;
                switch (key) {
                    case 'next':
                        changeFocusItem((_a = vsRef.current) === null || _a === void 0 ? void 0 : _a.scrollToStep(1));
                        break;
                    case 'prev':
                        changeFocusItem((_b = vsRef.current) === null || _b === void 0 ? void 0 : _b.scrollToStep(-1));
                        break;
                    case 'first':
                        changeFocusItem((_c = vsRef.current) === null || _c === void 0 ? void 0 : _c.scrollToStart());
                        break;
                    case 'last':
                        changeFocusItem((_d = vsRef.current) === null || _d === void 0 ? void 0 : _d.scrollToEnd());
                        break;
                    default:
                        break;
                }
            } }, { children: ({ render: renderComboboxKeyboard }) => {
                const input = renderComboboxKeyboard(cloneHTMLElement(el, {
                    value: searchValue,
                    'aria-controls': listId,
                    onBlur: (e) => {
                        var _a, _b;
                        (_b = (_a = el.props).onBlur) === null || _b === void 0 ? void 0 : _b.call(_a, e);
                        changeVisible(false);
                    },
                    onKeyDown: (e) => {
                        var _a, _b;
                        (_b = (_a = el.props).onKeyDown) === null || _b === void 0 ? void 0 : _b.call(_a, e);
                        if ((e.code === 'Enter' || (!dSearchable && e.code === 'Space')) && visible && focusItem) {
                            e.preventDefault();
                            if (focusItem[IS_CREATE]) {
                                createItem(focusItem);
                            }
                            changeSelectByClick(focusItem.value);
                        }
                    },
                    onChange: (e) => {
                        var _a, _b;
                        (_b = (_a = el.props).onChange) === null || _b === void 0 ? void 0 : _b.call(_a, e);
                        const val = e.currentTarget.value;
                        if (dSearchable) {
                            changeSearchValue(val);
                        }
                    },
                }));
                return isUndefined(dInputRender) ? input : dInputRender(input);
            } }))), onFocusVisibleChange: setFocusVisible, onClear: () => {
            onClear === null || onClear === void 0 ? void 0 : onClear();
            if (dMultiple) {
                changeSelect([]);
            }
            else {
                changeSelect(null);
            }
        } }, { children: ({ renderPopup }) => (_jsx(DTransition, Object.assign({ dIn: visible, dDuring: TTANSITION_DURING_POPUP, onEnter: updatePosition, afterEnter: () => {
                afterVisibleChange === null || afterVisibleChange === void 0 ? void 0 : afterVisibleChange(true);
            }, afterLeave: () => {
                afterVisibleChange === null || afterVisibleChange === void 0 ? void 0 : afterVisibleChange(false);
            } }, { children: (state) => {
                let transitionStyle = {};
                switch (state) {
                    case 'enter':
                        transitionStyle = { transform: 'scaleY(0.7)', opacity: 0 };
                        break;
                    case 'entering':
                        transitionStyle = {
                            transition: ['transform', 'opacity'].map((attr) => `${attr} ${TTANSITION_DURING_POPUP}ms ease-out`).join(', '),
                            transformOrigin,
                        };
                        break;
                    case 'leaving':
                        transitionStyle = {
                            transform: 'scaleY(0.7)',
                            opacity: 0,
                            transition: ['transform', 'opacity'].map((attr) => `${attr} ${TTANSITION_DURING_POPUP}ms ease-in`).join(', '),
                            transformOrigin,
                        };
                        break;
                    case 'leaved':
                        transitionStyle = { display: 'none' };
                        break;
                    default:
                        break;
                }
                return renderPopup(_jsxs("div", Object.assign({ ref: popupRef, className: getClassName(dPopupClassName, `${dPrefix}select__popup`), style: Object.assign(Object.assign({}, popupPositionStyle), transitionStyle) }, { children: [dLoading && (_jsx("div", Object.assign({ className: getClassName(`${dPrefix}select__loading`, {
                                [`${dPrefix}select__loading--empty`]: list.length === 0,
                            }) }, { children: _jsx(LoadingOutlined, { dSize: list.length === 0 ? 18 : 24, dSpin: true }) }))), _jsx(DVirtualScroll, Object.assign({}, vsPerformance, { ref: vsRef, dFillNode: _jsx("li", {}), dItemRender: (item, index, { aria, vsList }, parent) => {
                                const { label: itemLabel, value: itemValue, disabled: itemDisabled, children } = item;
                                const node = dCustomItem ? dCustomItem(item) : itemLabel;
                                if (children) {
                                    return (_jsxs("ul", Object.assign({ className: `${dPrefix}select__option-group`, role: "group", "aria-labelledby": getGroupId(itemValue) }, { children: [_jsx("li", Object.assign({ id: getGroupId(itemValue), className: `${dPrefix}select__option-group-label`, role: "presentation" }, { children: _jsx("div", Object.assign({ className: `${dPrefix}select__option-content` }, { children: node })) }), itemValue), vsList] }), itemValue));
                                }
                                let isSelected = false;
                                if (dMultiple) {
                                    isSelected = select.has(itemValue);
                                }
                                else {
                                    isSelected = select === itemValue;
                                }
                                return (_createElement("li", Object.assign({}, aria, { key: itemValue, id: getItemId(itemValue), className: getClassName(`${dPrefix}select__option`, {
                                        'is-selected': !dMultiple && isSelected,
                                        'is-disabled': itemDisabled,
                                    }), style: { paddingLeft: parent.length === 0 ? undefined : 12 + 8 }, title: (item[IS_CREATE] ? t('Create') + ' ' : '') + itemLabel, role: "option", "aria-selected": isSelected, "aria-disabled": itemDisabled, onClick: () => {
                                        if (!itemDisabled) {
                                            if (item[IS_CREATE]) {
                                                createItem(item);
                                            }
                                            changeFocusItem(item);
                                            changeSelectByClick(itemValue);
                                        }
                                    } }),
                                    focusVisible && (focusItem === null || focusItem === void 0 ? void 0 : focusItem.value) === itemValue && _jsx("div", { className: `${dPrefix}focus-outline` }),
                                    item[IS_CREATE] ? (_jsx(PlusOutlined, { dTheme: "primary" })) : dMultiple ? (_jsx(DCheckbox, { dDisabled: itemDisabled, dModel: isSelected })) : null,
                                    _jsx("div", Object.assign({ className: `${dPrefix}select__option-content` }, { children: node }))));
                            }, dFocusItem: focusItem, dSize: dVirtual ? 264 : Infinity, dPadding: 4, dEmptyRender: () => (_jsx("li", Object.assign({ className: `${dPrefix}select__empty`, style: { paddingLeft: 12 + 8 } }, { children: _jsx("div", Object.assign({ className: `${dPrefix}select__option-content` }, { children: t('No Data') })) }))), onScrollEnd: onScrollBottom }, { children: ({ render, vsList }) => render(_jsx("ul", Object.assign({ id: listId, className: `${dPrefix}select__list`, style: { pointerEvents: dLoading ? 'none' : undefined }, tabIndex: -1, role: "listbox", "aria-multiselectable": dMultiple, "aria-activedescendant": isUndefined(focusItem) ? undefined : getItemId(focusItem.value) }, { children: list.length === 0 ? (_jsx("li", Object.assign({ className: `${dPrefix}select__empty` }, { children: _jsx("div", Object.assign({ className: `${dPrefix}select__option-content` }, { children: t('No Data') })) }))) : (vsList) }))) }))] })));
            } }))) })));
}
export const DSelect = React.forwardRef(Select);
